/**************************************************************************************************
 * (C) Copyright 2004, ELSTA Sp. z o.o., Wieliczka, Poland,  http://www.elsta.pl
 *
 * Author: Artur Lipowski
 * Product Manager: Cezary Worek
 *
 * No guarantees, warrantees, or promises, implied or otherwise.
 * May be used for hobby or commercial purposes provided copyright
 * notice remains intact. 
 * 
 * $RCSfile: arm.h,v $
 * $Revision: 1.6 $
 * $Date: 2004/04/22 19:45:56 $
 *
 *************************************************************************************************/

#ifndef ARM_H
#define ARM_H


//
// (S)CPSR Register defines
//
 
#define CPSR_IRQ_DISABLE        0x80
#define CPSR_FIQ_DISABLE        0x40
#define CPSR_THUMB_ENABLE       0x20
#define CPSR_USER_MODE          0x10
#define CPSR_FIQ_MODE           0x11
#define CPSR_IRQ_MODE           0x12
#define CPSR_SUPERVISOR_MODE    0x13
#define CPSR_UNDEF_MODE         0x1B
#define CPSR_SYSTEM_MODE        0x1F

#define CPSR_MODE_BITS          0x1F


#ifndef __ASSEMBLER__

#define SOFTINT_NOARG(_fn_)	asm volatile ("swi " #_fn_)

#define SOFTINT_SET(_fn_, _arg_)	\
	asm volatile (			\
		"mov r0, %0	\n\t"\
		"swi " #_fn_ " \n\t"\
		:		\
		: "r"(_arg_)	\
		: "r0"		\
	);

#define SOFTINT_SET_TWO(_fn_, _arg1_, _arg2_)	\
	asm volatile (			\
		"mov r0, %0	\n\t"		\
		"mov r1, %1	\n\t"		\
		"swi " #_fn_ " \n\t"		\
		:				\
		: "r"(_arg1_), "r"(_arg2_)	\
		: "r0", "r1"			\
	);

#define SOFTINT_GET(_fn_, _arg_)	\
	asm volatile (			\
		"swi " #_fn_ " \n\t"\
		"mov %0, r0	\n\t"\
		: "=r"(_arg_)	\
		:		\
		: "r0"		\
	);

#define SOFTINT_GET_TWO(_fn_, _arg1_, _arg2_)	\
	asm volatile (			\
		"swi " #_fn_ " \n\t"		\
		"mov %0, r0	\n\t"		\
		"mov %1, r1	\n\t"		\
		: "=r"(_arg1_), "=r"(_arg2_)	\
		:				\
		: "r0", "r1"			\
	);


#ifndef __thumb__

#define DISABLE_INTS(_old_)\
    asm volatile (		\
        "mrs %0,CPSR		\n\t"\
        "mrs r4,CPSR		\n\t"\
        "orr r4,r4,#CPSR_IRQ_DISABLE	\n\t"\
        "msr CPSR_c,r4		\n\t"\
        : "=r"(_old_)		\
        :			\
        : "r4"			\
        );

#define ENABLE_INTS()	\
    asm volatile (		\
        "mrs r3,CPSR		\n\t"\
        "bic r3,r3,#CPSR_IRQ_DISABLE	\n\t"\
        "msr CPSR_c,r3		\n\t"\
        :			\
        :			\
        : "r3"			\
        );

#define RESTORE_INTS(_old_)\
    asm volatile (		\
        "mrs r3,CPSR		\n\t"\
        "and r4,%0,#CPSR_IRQ_DISABLE	\n\t"\
        "bic r3,r3,#CPSR_IRQ_DISABLE	\n\t"\
        "orr r3,r3,r4		\n\t"\
        "msr CPSR_c,r3		\n\t"\
        :			\
        : "r"(_old_)		\
        : "r3", "r4"		\
        );

#define QUERY_INTS(_old_)	\
    asm volatile (		\
        "mrs r4,CPSR		\n\t"\
        "and r4,r4,#CPSR_IRQ_DISABLE	\n\t"\
        "eor %0,r4,#CPSR_IRQ_DISABLE	\n\t"\
        : "=r"(_old_)		\
        :			\
        : "r4"			\
        );
#else // __thumb__

// Thumb mode does not have access to the PSR registers;  

#define SWITCH_TO_ARM \
        "add r4, pc, #4		\n\t"\
        "bx r4			\n\t"\
        "nop			\n\t"\
        ".align 2		\n\t"\
        ".arm			\n\t"

#define SWITCH_TO_THUMB \
        "add r4, pc, #1		\n\t"\
        "bx r4			\n\t"\
        ".thumb_func		\n\t"\
        "L_thumb_mode_%=:	"

#define DISABLE_INTS(_old_) \
    asm volatile (		\
	SWITCH_TO_ARM		\
        "mrs %0, CPSR		\n\t"\
        "mrs r4, CPSR		\n\t"\
	"L_again_%=:		\n\t"\
        "orr r4, r4, #0x80	\n\t"\
        "msr CPSR_c, r4		\n\t"\
        "mrs r4, CPSR		\n\t"\
        "tst r4, #0x80	\n\t"\
        "beq L_again_%=		\n\t"\
	SWITCH_TO_THUMB 	\
        : "=r"(_old_)		\
        :			\
	: "r4"			\
        );

#define ENABLE_INTS() \
    asm volatile (		\
	SWITCH_TO_ARM		\
        "mrs r4, CPSR		\n\t"\
        "bic r4, r4, #0x80	\n\t"\
        "msr CPSR_c, r4		\n\t"\
	SWITCH_TO_THUMB 	\
        :			\
        :			\
	: "r4"			\
        );			\

#define RESTORE_INTS(_old_)	\
    asm volatile (		\
	SWITCH_TO_ARM		\
	"L_again_%=:		\n\t"\
        "mrs r3, CPSR		\n\t"\
        "tst %0, #0x80	\n\t"\
        "biceq r3, r3, #0x80	\n\t"\
        "orrne r3, r3, #0x80	\n\t"\
        "msr CPSR_c, r3		\n\t"\
        "mrs r3, CPSR		\n\t"\
        "and r4, %0, #0x80	\n\t"\
        "and r3, r3, #0x80	\n\t"\
        "eors r4, r4, r3	\n\t"\
        "bne L_again_%=		\n\t"\
	SWITCH_TO_THUMB 	\
        :			\
        : "r"(_old_)		\
	: "r3", "r4"		\
        );


#define QUERY_INTS(_old_) \
    asm volatile (		\
	SWITCH_TO_ARM		\
	"mrs r4, CPSR		\n\t"\
	"and %0, r4, #0x80	\n\t"\
	SWITCH_TO_THUMB 	\
	: "=r"(_old_)		\
	:			\
	: "r4"			\
	);

#endif // __thumb__

#endif	// __ASSEMBLER__

#endif // ARM_H
