/**************************************************************************************************
 * (C) Copyright 2004, ELSTA Sp. z o.o., Wieliczka, Poland,  http://www.elsta.pl
 *
 * Author: Artur Lipowski
 * Product Manager: Cezary Worek
 *
 * No guarantees, warrantees, or promises, implied or otherwise.
 * May be used for hobby or commercial purposes provided copyright
 * notice remains intact. 
 * 
 * $RCSfile: macros.S,v $
 * $Revision: 1.12 $
 * $Date: 2004/04/25 07:23:58 $
 *
 *************************************************************************************************/

#define PTR(name)	.##name: .word  name

.macro THUMB_MODE _r_
	.arm
	add	\_r_, pc, #1
	bx	\_r_
	.thumb
	.thumb_func
9:
 .endm

.macro ARM_MODE _r_
	.thumb
	add	\_r_, pc, #4
	bx	\_r_
	nop
	.align 2
	.arm
.endm


.macro THUMB_CALL_FAR_THUMB _sub_, _r_
        .thumb
	ldr     \_r_, =2f+1
	mov	lr, \_r_
	ldr	\_r_, =\_sub_
	mov	pc, \_r_
	.align 2
        .thumb
        .thumb_func
2:
.endm

.macro ARM_CALL_FAR_THUMB _sub_, _r_
	.arm
	ldr	\_r_, =\_sub_
	add	lr, pc, #1
	bx	\_r_
	.thumb
        .thumb_func
.endm

.macro ARM_CALL_FAR_ARM _sub_
        mov     lr, pc
	ldr	pc, =\_sub_
.endm

//
//	Task frame:
//
//	address	
//	offset	content
//
//	 0:	task CPSR
//	 4:	task restart address ( lr_[svc]|irq] )
//	 8:	R0
//	12:	R1
//	16:	R2
//	20:	R3
//	24:	R4
//	28:	R5
//	32:	R6
//	36:	R7
//	40:	R8
//	44:	R9
//	48:	R10
//	52:	R11
//	56:	R12
//	60:	R13_usr
//	64:	R14_usr
//	

.macro SW_INT_ENTER
	stmfd	sp!, {r0}
.endm

.macro SW_INT_EXIT
	ldmfd	sp!, {r0}
	movs	pc, lr
.endm

.macro SWI_SAVE_FRAME
	ldr	r0, =OSTCBCur		// Get current task TCB address
	ldr	r0, [r0]		// OSTCBCur->OSTCBStkPtr = frame address
	ldr	r0, [r0]
	add	r0, r0, #12		// seek to r1 location
	stmia	r0, {r1-r14}^
	nop
	mrs	r1, SPSR
	mov	r2, lr
	ldmfd	sp!, {r3}		// pop r0
	sub	r0, r0, #12		// seek to CPSR location
	stmia	r0, {r1-r3}
.endm

.macro INT_ENTER
	stmfd   sp!, {r12, lr}
	mrs	r12, SPSR
	stmfd	sp!, {r0-r3, r12}
.endm

.macro INT_EXIT
	ldmfd	sp!, {r0-r3, r12}
	msr	SPSR_fsxc, r12
	ldmfd	sp!, {r12, pc}^
.endm

//	SP
//	offset	content
//	0	r0
//	4	r1
//	8	r2
//	12	r3
//	16	SPSR
//	20	r12
//	24	lr

.macro SAVE_FRAME_SKIP_SP words_to_skip
	ldr	r0, =OSTCBCur		// Get current task TCB address
	ldr	r0, [r0]		// OSTCBCur->OSTCBStkPtr = frame address
	ldr	r0, [r0]
	add	r0, r0, #8		// seek to r0
	add	r1, sp, #\words_to_skip*4	// calculate top of the stack
	ldmfd	r1!, {r2, r3, r12, r14}	// pop r0 - r3
	stmia	r0!, {r2, r3, r12, r14}
	ldmfd	r1, {r3, r12, r14}	// pop SPSR, r12, lr
	str	r3, [r0, #-24]		// store SPSR
	str	lr, [r0, #-20]		// store lr
	stmia	r0, {r4-r14}^
	nop
.endm

.macro LOAD_FRAME _ostbcur_
	ldr	r0, [\_ostbcur_]	// OSTCBCur->OSTCBStkPtr = frame address
	ldmia	r0!, {r12, lr}
	msr	SPSR_fsxc, r12
	ldmia	r0, {r0-r14}^
	nop
	movs	pc, lr
.endm


.macro SWITCH_TO_MODE _mode_, _r_
	mrs     \_r_, CPSR
	bic     \_r_, \_r_, #CPSR_MODE_BITS
	orr     \_r_, \_r_, #\_mode_
	msr     CPSR_c, \_r_
.endm

.macro SWITCH_TO_MODE_DI _mode_, _r_
	mrs     \_r_, CPSR		// Disable Interrupt and switch back in mode
	bic     \_r_, \_r_, #CPSR_MODE_BITS
	orr     \_r_, \_r_, #CPSR_IRQ_DISABLE|\_mode_
	msr     CPSR_c, \_r_
.endm

.macro SWITCH_TO_MODE_EI _mode_, _r_
	mrs     \_r_, CPSR		// Disable Interrupt and switch back in mode
	bic     \_r_, \_r_, #CPSR_MODE_BITS|CPSR_IRQ_DISABLE
	orr     \_r_, \_r_, #\_mode_
	msr     CPSR_c, \_r_
.endm

.macro SWITCH_TO_SYSMODE_EI _r_
	mrs	\_r_, SPSR		// conditionally enable interrupts
	tst	\_r_, #CPSR_IRQ_DISABLE
        mrs	\_r_, CPSR        	// switch to SYS Mode
	biceq   \_r_, \_r_, #CPSR_IRQ_DISABLE
	orr     \_r_, \_r_, #CPSR_SYSTEM_MODE
	msr     CPSR_c, \_r_
.endm

