/**************************************************************************************************
 * (C) Copyright 2004, ELSTA Sp. z o.o., Wieliczka, Poland,  http://www.elsta.pl
 *
 * Author: Artur Lipowski
 * Product Manager: Cezary Worek
 *
 * No guarantees, warrantees, or promises, implied or otherwise.
 * May be used for hobby or commercial purposes provided copyright
 * notice remains intact. 
 * 
 * $RCSfile: startup.S,v $
 * $Revision: 1.17 $
 * $Date: 2004/04/22 19:52:59 $
 *
 *************************************************************************************************/

#include <arm.h>
#include <at91_io.h>
#include "macros.S"

	.pool
	.code   32
	.section .startup, "ax", %progbits

	.global reset_vector
reset_vector:

	ldr	r1, _init_flag
	cmp	r1, #0
dead_end:
	bne dead_end				// If we get restarted, hang here 
						// to avoid corrupting memory
	ldr	r1, =0xDEAD00
	str	r1, _init_flag
	
	// needed only if redboot is used
	mrs	r0, CPSR
	orr	r0, r0, #CPSR_IRQ_DISABLE|CPSR_FIQ_DISABLE
	msr	CPSR_c, r0

	ldr	r1, =0xFFFFFFFF
	ldr	r0, =AT91_PS			// Power saving interface
	str	r1, [r0, #AT91_PS_PCER]		// Enable all peripheral [clocks]

	ldr	r0, =AT91_AIC
	str	r1, [r0, #AT91_AIC_IDCR]	// Reset all interrupts
	ldr	r2, =8
1:
	str	r1, [r0, #AT91_AIC_EOI]		// Flush internal priority level stack
	sub	r2, r2, #1
	bne	1b
	
	ldr	r1, =at91_spurious_handler	// Set up the default vectors
	str	r1, [r0, #AT91_AIC_SPU]

	ldr	r1, =at91_default_fiq_handler
	str	r1, [r0, #AT91_AIC_SVR]		// FIQ vector at slot #0

	ldr	r1, =at91_default_irq_handler
	add	r0, r0, #AT91_AIC_SVR
	mov	r2, #31
1:
	str	r1, [r0, r2, LSL #2]
	subs	r2, r2, #1			// omit FIQ
	bhi	1b

	ldr	r0, =AT91_PIO		// disable PIO, so peripherals can use: 
	ldr	r1, =AT91_PIO_MUXPINS	// UART, FIQ, EINT, Timer clocks
	str	r1, [r0, #AT91_PIO_PDR]

	// reloacte .vectors setion to address 0
	ldr	r10, ._vectors_src	// source
	ldr	r11, ._v_start		// destination
	ldr	r12, ._v_end		// up to
1:
	ldmia	r10!, {r0-r9}
	stmia	r11!, {r0-r9}
	cmp	r11, r12
	blo	1b

	ldr	r0, =0			// clear BSS
	ldr	r1, =0			// set (5*4) bytes at once
	ldr	r2, =0
	ldr	r3, =0
	ldr	r4, =0
	ldr	r5, .__bss_start__
	ldr	r6, .__bss_end__
	cmp	r5, r6
	bhs	2f
1:
	stmia	r5!, {r0-r4}
	cmp	r5, r6
	blo	1b
2:

	// setup stack pointers for CPU modes
	mrs	r0, CPSR

	bic	r0, r0, #CPSR_MODE_BITS
	orr	r0, r0, #CPSR_FIQ_MODE
	msr	CPSR_c, r0
	ldr	sp, ._fiq_stack

	ldr	r1, ._int_stack
	bic	r0, r0, #CPSR_MODE_BITS
	orr	r0, r0, #CPSR_IRQ_MODE
	msr	CPSR_c, r0
	mov	sp, r1

	sub	r1, r1, #100*4	// allocate IRQ mode stack

	bic	r0, r0, #CPSR_MODE_BITS
	orr	r0, r0, #CPSR_SUPERVISOR_MODE
	msr	CPSR_c, r0
	mov	sp, r1

	bic	r0, r0, #CPSR_MODE_BITS
	orr	r0, r0, #CPSR_SYSTEM_MODE
	msr	CPSR_c, r0
	ldr	sp, ._main_stack		// stay in system mode
 	
	// Run kernel + application in THUMB or ARM mode (depends on the _main address)
	// Jump to main function
	ldr	r0, =_main
	bx	r0


PTR(__bss_start__)
PTR(__bss_end__)
PTR(_main_stack)
PTR(_int_stack)
PTR(_fiq_stack)
PTR(_vectors_src)
PTR(_v_start)
PTR(_v_end)
_init_flag:	.long   0

	.end

