/**************************************************************************************************
 * (C) Copyright 2004, ELSTA Sp. z o.o., Wieliczka, Poland,  http://www.elsta.pl
 *
 * Author: Artur Lipowski
 * Product Manager: Cezary Worek
 *
 * No guarantees, warrantees, or promises, implied or otherwise.
 * May be used for hobby or commercial purposes provided copyright
 * notice remains intact. 
 * 
 * $RCSfile: os_cpu_a.S,v $
 * $Revision: 1.12 $
 * $Date: 2004/04/25 07:09:42 $
 *
 *************************************************************************************************/

#include <arm.h>
#include <at91_io.h>
#include <macros.S>
#include "os_cfg.h"

	.pool
	.arm
	.section .vectors, "ax", %progbits

//--------------------------------------------------------------------------------------
//
// system_timer_ISR
//
//--------------------------------------------------------------------------------------
	.global system_timer_ISR
system_timer_ISR:
	sub	lr, lr, #4		// Adjust LR
	INT_ENTER

	ldr	r1, =AT91_TC_TC0	// Read status register to reenable
	ldr	r0, [r1, #AT91_TC_SR]	// CPC flag change -> generate interrupt

	ldr	r1, =OSIntNesting	// increment OSIntNesting
	ldrb	r0, [r1]
	add	r0, r0, #1
	strb	r0, [r1]

	cmp	r0, #1			// if OSIntNesting == 1 then
	bne	1f			// 	save task frame at the task stack
	SAVE_FRAME_SKIP_SP 0
1:

#ifdef ENABLE_NESTED_INTS
	SWITCH_TO_MODE_EI CPSR_SYSTEM_MODE, r0
#else
	SWITCH_TO_MODE CPSR_SYSTEM_MODE, r0
#endif

	stmfd	sp!, {lr}
#ifndef OS_IN_ARM_MODE
	ARM_CALL_FAR_THUMB OSTimeTick, r0
	THUMB_CALL_FAR_THUMB OSIntExit, r0
	ARM_MODE r0
#else
	ARM_CALL_FAR_ARM OSTimeTick
	ARM_CALL_FAR_ARM OSIntExit
#endif	// OS_IN_ARM_MODE
	ldmfd	sp!, {lr}

	SWITCH_TO_MODE_DI CPSR_IRQ_MODE, r0

	ldr	r0, =AT91_AIC		//  Mark the End of Interrupt
	str	r0, [r0, #AT91_AIC_EOI]

	INT_EXIT


//--------------------------------------------------------------------------------------
//
// software_interrupt
//
//--------------------------------------------------------------------------------------
#define _LED_GREEN AT91_PIO_PSR_TIOB0
	.global software_interrupt
software_interrupt:
	SW_INT_ENTER
	mrs	r0, SPSR
	tst	r0, #CPSR_THUMB_ENABLE
	ldrneh	r0, [lr, #-2]
	ldreq	r0, [lr, #-4]
	and	r0, r0, #0xFF
	cmp	r0, #5
	ldrls	pc, [pc, r0, lsl #2]
	b	end_sw_irq

	.word	fn_0
	.word	fn_1
	.word	fn_2
	.word	fn_3
	.word	fn_4
	.word	fn_5

//-------------------------------------------------------------
fn_0:				// context switch
//-------------------------------------------------------------
	SWI_SAVE_FRAME

#if OS_CPU_HOOKS_EN > 0 && OS_TASK_SW_HOOK_EN > 0
	SWITCH_TO_MODE CPSR_SYSTEM_MODE, r0

	// we not need save user registers because they are saved already
#ifndef OS_IN_ARM_MODE
	ARM_CALL_FAR_THUMB OSTaskSwHook, r0
	ARM_MODE r0
#else
	ARM_CALL_FAR_ARM OSTaskSwHook
#endif	// OS_IN_ARM_MODE

	SWITCH_TO_MODE_DI CPSR_SUPERVISOR_MODE, r0
#endif

	ldr	r0, =OSPrioHighRdy
	ldrb	r1, [r0]
	ldr	r0, =OSPrioCur
	strb	r1, [r0]	// OSPrioCur = OSPrioHighRdy

	ldr	r0, =OSTCBHighRdy
	ldr	r1, [r0]
	ldr	r0, =OSTCBCur
	str	r1, [r0]	// OSTCBCur = OSTCBHighRdy

	LOAD_FRAME r1

//-------------------------------------------------------------
fn_1:				// disable interrupts
//-------------------------------------------------------------
	mrs	r0, SPSR
	str	r0, [sp]
	orr	r0, r0, #CPSR_IRQ_DISABLE
	msr	SPSR_c, r0
	SW_INT_EXIT

//-------------------------------------------------------------
fn_2:				// enable interrupts
//-------------------------------------------------------------
	mrs	r0, SPSR
	str	r0, [sp]
	bic	r0, r0, #CPSR_IRQ_DISABLE
	msr	SPSR_c, r0
	SW_INT_EXIT

//-------------------------------------------------------------
fn_3:				// restore interrupts
//-------------------------------------------------------------
	ldr	r0, [sp]
	tst	r0, #CPSR_IRQ_DISABLE
	mrs	r0, SPSR
	biceq	r0, r0, #CPSR_IRQ_DISABLE
	orrne	r0, r0, #CPSR_IRQ_DISABLE
	msr	SPSR_c, r0
	SW_INT_EXIT

//-------------------------------------------------------------
fn_4:				// get system RTC
//-------------------------------------------------------------
	stmfd	sp!, {lr}
	bl	_get_sys_RTC_vars
	ldmfd	sp!, {lr}
	// here we have seconds_epoch in r0 and ot_seconds in r1
	str	r0, [sp]
	SW_INT_EXIT

//-------------------------------------------------------------
fn_5:				// set system RTC
//-------------------------------------------------------------
	ldr	r0, [sp]
	// here we have seconds_epoch in r0 and ot_seconds in r1
	stmfd	sp!, {lr}
	bl	_set_sys_RTC_vars
	ldmfd	sp!, {lr}
/*
	mov	r0, #_LED_GREEN
	ldr	r1, =AT91_PIO
	str	r0, [r1, #AT91_PIO_SODR]	// turn on green LED
1:
	b	1b
*/

end_sw_irq:
	SW_INT_EXIT



//--------------------------------------------------------------------------------------
//
// .text section
//
//--------------------------------------------------------------------------------------
	.pool
#ifndef OS_IN_ARM_MODE
	.thumb
#else
	.arm
#endif	// OS_IN_ARM_MODE
	.text

//--------------------------------------------------------------------------------------
//
// OSStartHighRdy
//
//--------------------------------------------------------------------------------------
	.global OSStartHighRdy
#ifndef OS_IN_ARM_MODE
	.thumb_func
#else
#endif	// OS_IN_ARM_MODE

OSStartHighRdy:

#if OS_CPU_HOOKS_EN > 0 && OS_TASK_SW_HOOK_EN > 0
	bl	OSTaskSwHook
#endif

#ifndef OS_IN_ARM_MODE
	ARM_MODE r1
#else
#endif	// OS_IN_ARM_MODE

	ldr	r0, =OSRunning
	mov	r1, #1
	strb	r1, [r0]		// OSRunning = TRUE

	ldr	r0, =OSTCBHighRdy	// Get highest priority task TCB address
	ldr	r0, [r0]
	ldr	r0, [r0]		// get new task's frame pointer address

	ldmia	r0!, {r12, lr}
#ifndef OS_IN_ARM_MODE
	// because we need access to CPSR to stay in ARM mode
	bic     r12, r12, #CPSR_THUMB_ENABLE
#else
#endif
	msr	CPSR_c, r12
	ldmia	r0, {r0-r13}
	nop
#ifndef OS_IN_ARM_MODE
	bx	lr
#else
	mov	pc, lr
#endif	// OS_IN_ARM_MODE

//--------------------------------------------------------------------------------------
//
// OSIntCtxSw
//
//--------------------------------------------------------------------------------------
	.global OSIntCtxSw
#ifndef OS_IN_ARM_MODE
	.thumb_func
#endif	// OS_IN_ARM_MODE
OSIntCtxSw:

#if OS_CPU_HOOKS_EN > 0 && OS_TASK_SW_HOOK_EN > 0
	bl	OSTaskSwHook
#endif

#ifndef OS_IN_ARM_MODE
	ARM_MODE r0
#else
#endif

	// we do not need to restore lr_sys because 
	// it will be reloaded from the frame of the new task

	SWITCH_TO_MODE_DI CPSR_IRQ_MODE, r0

	ldr	r0, =OSPrioHighRdy
	ldrb	r1, [r0]
	ldr	r0, =OSPrioCur
	strb	r1, [r0]		// OSPrioCur = OSPrioHighRdy

	add	sp, sp, #7*4		// remove from the stack regs of the previous task

	ldr	r0, =AT91_AIC		//  Mark the End of Interrupt
	str	r0, [r0, #AT91_AIC_EOI]

	ldr	r0, =OSTCBHighRdy
	ldr	r1, [r0]
	ldr	r0, =OSTCBCur
	str	r1, [r0]		// OSTCBCur = OSTCBHighRdy

	LOAD_FRAME r1


	.end
