/**************************************************************************************************
 * (C) Copyright 2004, ELSTA Sp. z o.o., Wieliczka, Poland,  http://www.elsta.pl
 *
 * Author: Artur Lipowski
 * Product Manager: Cezary Worek
 *
 * No guarantees, warrantees, or promises, implied or otherwise.
 * May be used for hobby or commercial purposes provided copyright
 * notice remains intact. 
 * 
 * $RCSfile: vectors.S,v $
 * $Revision: 1.18 $
 * $Date: 2004/05/04 11:08:22 $
 *
 *************************************************************************************************/

#include <arm.h>
#include <at91_io.h>
#include "macros.S"

#define _LED_RED AT91_PIO_PSR_TIOA0
#define _LED_AMBER AT91_PIO_PSR_TIOA1 
#define _LED_GREEN AT91_PIO_PSR_TIOB0

#define _ALL_LEDS _LED_GREEN + _LED_RED + _LED_AMBER 

#define _FIQ_CODE _LED_RED
#define _IRQ_CODE _LED_AMBER
#define _UNDE_CODE _LED_RED + _LED_AMBER
#define _APRE_CODE _LED_GREEN
#define _ADAT_CODE _LED_GREEN + _LED_RED

	.pool
	.arm
	.section .vectors, "ax", %progbits

.org 0
	.global ISR_vector_table
ISR_vector_table:
	ldr	pc, =reset_vector	// 0x00
	b	undefined_instruction	// 0x04 undefined_instruction
	b	software_interrupt	// 0x08 start && software int
	b	abort_prefetch		// 0x0C abort_prefetch
	b	abort_data		// 0x10 abort_data
	.word	0			// unused
	ldr	pc, [pc, #-0xF20]	// 0x18 IRQ : read the AIC
	ldr	pc, [pc, #-0xF20]	// 0x1C FIQ : read the AIC


//--------------------------------------------------------------------------------------
//
// at91_default_fiq_handler
// at91_default_irq_handler
//
//--------------------------------------------------------------------------------------
	.global at91_default_fiq_handler
	.global at91_default_irq_handler
at91_default_fiq_handler:
	sub	lr, lr, #4
	mov	r5, #_FIQ_CODE
	b	blink_leds
at91_default_irq_handler:
	sub	lr, lr, #4
	mov	r5, #_IRQ_CODE
	b	blink_leds
undefined_instruction:
	sub	lr, lr, #4
	mov	r5, #_UNDE_CODE
	b	blink_leds
abort_prefetch:
	sub	lr, lr, #4
	mov	r5, #_APRE_CODE
	b	blink_leds
abort_data:
	sub	lr, lr, #8
	mov	r5, #_ADAT_CODE

blink_leds:
	mov	r6, lr
	mov	r7, #_ALL_LEDS
	ldr	r8, =AT91_PIO

1:
	str	r7, [r8, #AT91_PIO_CODR]	// turn off all LEDs
	str	r5, [r8, #AT91_PIO_SODR]	// display exception code
	mov	r0, #30
	bl	delay_ts			// delay 2s
	mov	r9, #0x80000000

2:
	str	r7, [r8, #AT91_PIO_CODR]	// turn off all LEDs
	mov	r0, #4
	bl	delay_ts			// delay 0.5s
	mov	r10, #_LED_GREEN
	tst	r6, r9
	addne	r10, r10, #_LED_RED		// turn on red LED if bit is 1
	str	r10, [r8, #AT91_PIO_SODR]	// display bit
	mov	r0, #8
	bl	delay_ts			// delay 1s
	movs	r9, r9, lsr #1
	bne	2b

	str	r7, [r8, #AT91_PIO_CODR]	// turn off all LEDs
	mov	r10, #_LED_AMBER
	str	r10, [r8, #AT91_PIO_SODR]	// turn on amber LED
	mov	r0, #30
	bl	delay_ts			// delay 2s
	b	1b

delay_ts:
1:
	mov	r1, #(ARM_AT91_CLOCK_SPEED/10)/4	// CPU clocks per 1/10 / inner loop clocks
2:
	nop
	subs	r1, r1, #1
	bne	2b
	subs	r0, r0, #1
	bne	1b
	mov	pc, lr

//--------------------------------------------------------------------------------------
//
// at91_spurious_handler
//
//--------------------------------------------------------------------------------------
	.global at91_spurious_handler
at91_spurious_handler:
	sub	lr, lr, #4			//  Adjust LR
	stmfd	sp!, {lr}			//  Save LR
	ldr	lr, =AT91_AIC			//  Mark the End of Interrupt
	str	lr, [lr, #AT91_AIC_EOI]
	ldmfd	sp!, {pc}^			//  Restore adjusted


	.end
